package com.tibco.contrib;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.apache.poi.ss.usermodel.Cell;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.streambase.liveview.client.LiveViewException;
import com.streambase.liveview.client.LiveViewExceptionType;
import com.streambase.liveview.client.LiveViewQueryLanguage;
import com.streambase.liveview.client.LiveViewQueryType;
import com.streambase.liveview.client.LiveViewTableCapability;
import com.streambase.liveview.client.Table.TableStatus;
import com.streambase.liveview.server.event.BeginSnapshotEvent;
import com.streambase.liveview.server.event.EndSnapshotEvent;
import com.streambase.liveview.server.event.EventListener;
import com.streambase.liveview.server.event.QueryExceptionEvent;
import com.streambase.liveview.server.event.TupleAddedEvent;
import com.streambase.liveview.server.query.QueryModel;
import com.streambase.liveview.server.table.CatalogedTable;
import com.streambase.liveview.server.table.Table;
import com.streambase.liveview.server.table.TablePublisher;
import com.streambase.liveview.server.table.plugin.TableProviderControl;
import com.streambase.sb.CompleteDataType;
import com.streambase.sb.Schema;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.util.Msg;

/*
 * This is the ExcelTable implementation. Each Excel file will be managed buy an instance of an ExcelTable
 * 
 */
public class ExcelTable implements Table  {

    private Logger logger = LoggerFactory.getLogger(ExcelTable.class);
    private BlockingQueue<ExcelTableItems> queue=new  LinkedBlockingQueue<ExcelTableItems>();
    AtomicBoolean shouldRun=new AtomicBoolean(true);
    
    // Some members for holding information about the Excel file 
	private POIFSFileSystem POIfs=null;
	private HSSFWorkbook workBook=null;
	private Schema schema=null;
	private CatalogedTable catalogTable=null;
	private int totalRows=0;
	private int numColumns=0;
	private HSSFSheet sheet;

	private final String excelPath;
	private final TableProviderControl helper;
	private File tableFile;
	private String tableName;
	
    public ExcelTable(TableProviderControl helper, String excelPath) {
    	this.excelPath=excelPath;
		this.helper=helper;
		
		// The work of parsing the Excel file and handling queries is all done in a separate thread. 
    	ExcelTableThread et=new ExcelTableThread(this);
    	et.start();
    }
    
	@Override
	public void addListener(EventListener listener, LiveViewQueryType arg1, QueryModel queryModel) throws LiveViewException {
		// Just add a new query to a work list
		queue.add(new ExcelTableItems(listener, queryModel));
	}

	@Override
	public TablePublisher createPublisher(String arg0) throws LiveViewException {
		throw LiveViewExceptionType.PUBLISHING_NOT_SUPPORTED.error();
	}

	@Override
	public void doDelete(String arg0, QueryModel arg1) throws LiveViewException {
		throw LiveViewExceptionType.TABLE_DOES_NOT_SUPPORT_QUERY.error();
	}

	@Override
	public QueryModel parseQuery(CatalogedTable catalogedTable, String queryString, LiveViewQueryType arg2, boolean arg3) throws LiveViewException {
        ExcelQueryModel model = new ExcelQueryModel(catalogedTable, this, queryString);
        // ExcelQueryModel does all the work
        model.validate(catalogedTable, false);
        return model;
	}

	@Override
	public void removeListener(EventListener arg0) throws LiveViewException {
		// Since this table provider only does snapshots, we don't have anything to clean up here.
	}

	public Schema getSchema() {
		return schema;
	}
	
	public String getTableName() {
		return tableName;
	}
	
	public void shutdown() {
		shouldRun.set(false);
	}
	
	/*
	 * Simple helper class to hold a query in the work list
	 */
	private class ExcelTableItems {
		
		public final EventListener listener;
		public final QueryModel queryModel;
		
		public ExcelTableItems(EventListener listener, QueryModel queryModel) {
			this.listener=listener;
			this.queryModel=queryModel;
		}
	}
	
	/*
	 * The ExcelTableThread does all the work of opening/parsing the Excel file,
	 * as well has actually servicing all queries.
	 */
	private class ExcelTableThread extends Thread {
		
		private final ExcelTable excelTable;
		
		public ExcelTableThread(ExcelTable et) {
			this.excelTable=et;
		}
		
		public void run() {
			
			try {
				// Find the Excel file 
				tableFile=helper.locateFile(excelPath, true);
				tableName = tableFile.getName();
				
				catalogTable = new CatalogedTable(tableName);
				catalogTable.setStatus(TableStatus.DISABLED, "Opening file");
				catalogTable.setCapabilities(EnumSet.of(LiveViewTableCapability.SNAPSHOT));
				catalogTable.setQueryLanguages( EnumSet.of(LiveViewQueryLanguage.OTHER));
				catalogTable.setDescription("TableProvider Excel Sample");
		        catalogTable.setCreateTime(Timestamp.now());
				
		        // Parse it
				POIfs = new POIFSFileSystem(new FileInputStream(tableFile));
				workBook = new HSSFWorkbook(POIfs);
			    sheet = workBook.getSheetAt(0);
			    HSSFRow row;
			    HSSFCell cell;

			    totalRows = sheet.getPhysicalNumberOfRows();
			    // This pre-supposes the data starts on the first row
			    numColumns = sheet.getRow(0).getPhysicalNumberOfCells();
			    
			    // Get the data types and create the Schema
			    row = sheet.getRow(0);
			    List<Field> fields = new ArrayList<Field>();
			    fields.add(new Field(ExcelTableProvider.ROW_COLUMN_NAME, CompleteDataType.forInt()));
	            for(Integer c = 0; c < numColumns; c++) {
	                cell = row.getCell(c);
	                if(cell != null) {
	                	// CELL_TYPE_BLANK, CELL_TYPE_BOOLEAN, CELL_TYPE_ERROR, CELL_TYPE_FORMULA, CELL_TYPE_NUMERIC, CELL_TYPE_STRING
	                	int type=cell.getCellType();
	                	String name= "C" + c.toString();
	                	switch (type) {
	                		case Cell.CELL_TYPE_NUMERIC:
	                			fields.add(new Field(name, CompleteDataType.forDouble()));
	                			break;
	                		case Cell.CELL_TYPE_BOOLEAN:
	                			fields.add(new Field(name, CompleteDataType.forBool()));
	                			break;
	                		case Cell.CELL_TYPE_STRING:
	                			fields.add(new Field(name, CompleteDataType.forString()));
	                			break;
	                		default:
	                			fields.add(new Field(name, CompleteDataType.forString()));
	                			break;
	                	}
	                }
	            }
	            
	            schema = new Schema(null, fields);
	            catalogTable.setSchema(schema);
	            List<String> pkey= new ArrayList<String>();
	            // For simplicity, Excel tables are given an extra column that is considered their primary key.
	            // The primary key is just the row number.
	            pkey.add(ExcelTableProvider.ROW_COLUMN_NAME);
	            catalogTable.setKeyFields(pkey);
	            
	            catalogTable.setRuntimeTable(excelTable);
	            helper.insert(catalogTable);
	            catalogTable.setStatus(TableStatus.ENABLED, "OK");
	            helper.upsert(catalogTable);

			} catch (Exception e) {
				logger.warn(Msg.format("Failed to open {0} - ({1})", tableName, e.getMessage()));
				catalogTable.setStatus(TableStatus.DISABLED, String.format("Failed to open %s (%s)", tableName, e.getMessage()));
			}
			
			// Now we are ready to start processing queries.
			while (shouldRun.get()) {
				ExcelTableItems ti=null;
				try {
					ti=null;
					ti=queue.poll(500, TimeUnit.MILLISECONDS);
					
					if (ti !=null) {
						ti.listener.sendEvent(new BeginSnapshotEvent(this));
						dumpRows(ti.listener, ti.queryModel);
					}
				} catch (LiveViewException le) {
					if (ti != null) {
						ti.listener.sendEvent(new QueryExceptionEvent(this, le));
					}
				} catch (Exception e) {
					logger.warn(Msg.format("Unexpected exception: {0}", e.getMessage()));
				} finally {
					// Even if dumpRows blows up we need to send an EndSnapshotEvent
					if (ti != null) {
						ti.listener.sendEvent(new EndSnapshotEvent(this));
					}
				}
			}
		}
		
		/*
		 * This is the method that reads the Excel data, serializes it into tuples and sends it to clients.
		 * 
		 * The "query language" is very primitive, it's simply <beginningRowNumber> [,<endRowNumber>]  
		 */
		private void dumpRows(EventListener listener, QueryModel queryModel) throws LiveViewException {
	        
	        int start=1;
	        int finish=totalRows;
	        String predicate = queryModel.getPredicate();
	        // Some clients like to send "true" when the user specifies no predicate
	        if ((predicate!=null) && !predicate.equalsIgnoreCase("true")) {
	        	String[] q=predicate.split(",");
		        start=Integer.parseInt(q[0]);
		        
		        if (q.length==2) {
		        	finish=Integer.parseInt(q[1]);
		        }
	        }
	        int limit=queryModel.getLimit();
	        
	        try {
			    HSSFRow row;
			    HSSFCell cell;
			    int rowCount=1;
			    for(int r = start; r <= finish; r++) {
			    	Tuple tuple = schema.createTuple();
			    	tuple.setInt(ExcelTableProvider.ROW_COLUMN_NAME, r);
			        
			    	// The query language enforces that rows start at one, but the excel lib starts at zero
			    	row = sheet.getRow(r-1);
			        if(row != null) {

			            for(int c = 0; c < numColumns; c++) {
			                cell = row.getCell(c);
			                // The tuple index is off by 1 because of ROW_COLUMN_NAME
			                int tupleIndex=c+1;
			                if(cell != null) {
		                		// build the tuple
		                       	int type=cell.getCellType();
		                       	try {
			                    	switch (type) {
			                    		case Cell.CELL_TYPE_NUMERIC:
			                    			tuple.setDouble(tupleIndex, cell.getNumericCellValue());
			                    			break;
			                    		case Cell.CELL_TYPE_BOOLEAN:
			                    			tuple.setBoolean(tupleIndex, cell.getBooleanCellValue());
			                    			break;
			                    		case Cell.CELL_TYPE_STRING:
			                    			tuple.setString(tupleIndex, cell.getStringCellValue().trim());
			                    			break;
			                    		default:
			                    			tuple.setString(tupleIndex, cell.getStringCellValue().trim());
			                    			break;
			                    	}
		                       	} catch (Exception e) {
		                       		logger.info(String.format("Error setting column %s with %s", tupleIndex, cell.toString()));
		                       	}
			                }
			            }
			            listener.sendEvent(new TupleAddedEvent(this, (long)r, tuple));
			        }
			        if(rowCount++ >= limit) {
			        	throw LiveViewExceptionType.QUERY_RESULT_QUEUE_SIZE_EXCEEDED.error(limit);
			        }
			    }
			    
			} catch (LiveViewException e) {
				throw e;
			} catch(Exception ioe) {
			    logger.info(Msg.format("Table {0}, failed to set tuple {1}", tableName, ioe.getMessage()));
			}
		}
	}
}
